@@ -1,55 +1,53 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * The Apache Software License, Version 1.1
+ *
+ * Copyright (c) 1996-2000 The Apache Software Foundation.  All rights
+ * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer. 
+ *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in
  *    the documentation and/or other materials provided with the
  *    distribution.
  *
- * 3. All advertising materials mentioning features or use of this
- *    software must display the following acknowledgment:
- *    "This product includes software developed by the Apache Group
- *    for use in the Apache HTTP server project (http://www.apache.org/)."
+ * 3. The end-user documentation included with the redistribution,
+ *    if any, must include the following acknowledgment:
+ *       "This product includes software developed by the
+ *        Apache Software Foundation (http://www.apache.org/)."
+ *    Alternately, this acknowledgment may appear in the software itself,
+ *    if and wherever such third-party acknowledgments normally appear.
  *
- * 4. The names "Apache Server" and "Apache Group" must not be used to
- *    endorse or promote products derived from this software without
- *    prior written permission.
+ * 4. The names "Apache" and "Apache Software Foundation" must
+ *    not be used to endorse or promote products derived from this
+ *    software without prior written permission. For written
+ *    permission, please contact apache@apache.org.
  *
- * 5. Redistributions of any form whatsoever must retain the following
- *    acknowledgment:
- *    "This product includes software developed by the Apache Group
- *    for use in the Apache HTTP server project (http://www.apache.org/)."
+ * 5. Products derived from this software may not be called "Apache",
+ *    nor may "Apache" appear in their name, without prior written
+ *    permission of the Apache Software Foundation.
  *
- * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
- * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
- * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
+ * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
+ * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
  * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
- * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
- * OF THE POSSIBILITY OF SUCH DAMAGE.
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
+ * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
+ * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
+ * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Apache Group and was originally based
- * on public domain software written at the National Center for
- * Supercomputing Applications, University of Illinois, Urbana-Champaign.
- * For more information on the Apache Group and the Apache HTTP server
- * project, please see <http://www.apache.org/>.
- *
  */
 
+
 /* 
  * And so it was decided the camel should be given magical multi-colored
  * feathers so it could fly and journey to once unknown worlds.
@@ -82,54 +80,59 @@ static HV *stacked_handlers = Nullhv;
 CPerlObj *pPerl;
 #endif
 
+typedef const char* (*crft)(); /* command_req_func_t */
+
 static command_rec perl_cmds[] = {
 #ifdef PERL_SECTIONS
-    { "<Perl>", perl_section, NULL, OR_ALL, RAW_ARGS, "Perl code" },
-    { "</Perl>", perl_end_section, NULL, OR_ALL, NO_ARGS, "End Perl code" },
+    { "<Perl>", (crft) perl_section, NULL, SECTION_ALLOWED, RAW_ARGS, "Perl code" },
+    { "</Perl>", (crft) perl_end_section, NULL, SECTION_ALLOWED, NO_ARGS, "End Perl code" },
 #endif
-    { "=pod", perl_pod_section, NULL, OR_ALL, RAW_ARGS, "Start of POD" },
-    { "=end", perl_pod_section, NULL, OR_ALL, RAW_ARGS, "End of =begin" },
-    { "=cut", perl_pod_end_section, NULL, OR_ALL, NO_ARGS, "End of POD" },
-    { "__END__", perl_config_END, NULL, OR_ALL, RAW_ARGS, "Stop reading config" },
-    { "PerlFreshRestart", perl_cmd_fresh_restart,
+    { "=pod", (crft) perl_pod_section, NULL, OR_ALL, RAW_ARGS, "Start of POD" },
+    { "=back", (crft) perl_pod_section, NULL, OR_ALL, RAW_ARGS, "End of =over" },
+    { "=cut", (crft) perl_pod_end_section, NULL, OR_ALL, NO_ARGS, "End of POD" },
+    { "__END__", (crft) perl_config_END, NULL, OR_ALL, RAW_ARGS, "Stop reading config" },
+    { "PerlFreshRestart", (crft) perl_cmd_fresh_restart,
       NULL,
       RSRC_CONF, FLAG, "Tell mod_perl to reload modules and flush Apache::Registry cache on restart" },
-    { "PerlTaintCheck", perl_cmd_tainting,
+    { "PerlTaintCheck", (crft) perl_cmd_tainting,
       NULL,
       RSRC_CONF, FLAG, "Turn on -T switch" },
 #ifdef PERL_SAFE_STARTUP
-    { "PerlOpmask", perl_cmd_opmask,
+    { "PerlOpmask", (crft) perl_cmd_opmask,
       NULL,
       RSRC_CONF, TAKE1, "Opmask File" },
 #endif
-    { "PerlWarn", perl_cmd_warn,
+    { "PerlWarn", (crft) perl_cmd_warn,
       NULL,
       RSRC_CONF, FLAG, "Turn on -w switch" },
-    { "PerlScript", perl_cmd_require,
+    { "PerlScript", (crft) perl_cmd_require,
       NULL,
       OR_ALL, ITERATE, "this directive is deprecated, use `PerlRequire'" },
-    { "PerlRequire", perl_cmd_require,
+    { "PerlRequire", (crft) perl_cmd_require,
       NULL,
       OR_ALL, ITERATE, "A Perl script name, pulled in via require" },
-    { "PerlModule", perl_cmd_module,
+    { "PerlModule", (crft) perl_cmd_module,
       NULL,
       OR_ALL, ITERATE, "List of Perl modules" },
-    { "PerlSetVar", perl_cmd_var,
+    { "PerlSetVar", (crft) perl_cmd_var,
       NULL,
       OR_ALL, TAKE2, "Perl config var and value" },
-    { "PerlSetEnv", perl_cmd_setenv,
+    { "PerlAddVar", (crft) perl_cmd_var,
+      (void*)1,
+      OR_ALL, ITERATE2, "Perl config var and value" },
+    { "PerlSetEnv", (crft) perl_cmd_setenv,
       NULL,
       OR_ALL, TAKE2, "Perl %ENV key and value" },
-    { "PerlPassEnv", perl_cmd_pass_env, 
+    { "PerlPassEnv", (crft) perl_cmd_pass_env, 
       NULL,
       RSRC_CONF, ITERATE, "pass environment variables to %ENV"},  
-    { "PerlSendHeader", perl_cmd_sendheader,
+    { "PerlSendHeader", (crft) perl_cmd_sendheader,
       NULL,
       OR_ALL, FLAG, "Tell mod_perl to parse and send HTTP headers" },
-    { "PerlSetupEnv", perl_cmd_env,
+    { "PerlSetupEnv", (crft) perl_cmd_env,
       NULL,
       OR_ALL, FLAG, "Tell mod_perl to setup %ENV by default" },
-    { "PerlHandler", perl_cmd_handler_handlers,
+    { "PerlHandler", (crft) perl_cmd_handler_handlers,
       NULL,
       OR_ALL, ITERATE, "the Perl handler routine name" },
 #ifdef PERL_TRANS
@@ -188,11 +191,11 @@ static handler_rec perl_handlers [] = {
 
 module MODULE_VAR_EXPORT perl_module = {
     STANDARD_MODULE_STUFF,
-    perl_startup,                 /* initializer */
+    perl_module_init,                 /* initializer */
     perl_create_dir_config,    /* create per-directory config structure */
     perl_merge_dir_config,     /* merge per-directory config structures */
     perl_create_server_config, /* create per-server config structure */
-    NULL,                      /* merge per-server config structures */
+    perl_merge_server_config,  /* merge per-server config structures */
     perl_cmds,                 /* command table */
     perl_handlers,             /* handlers */
     PERL_TRANS_HOOK,           /* translate_handler */
@@ -256,8 +259,6 @@ void perl_shutdown (server_rec *s, pool *p)
 
     if((pdl = getenv("PERL_DESTRUCT_LEVEL")))
 	perl_destruct_level = atoi(pdl);
-    else
-	perl_destruct_level = PERL_DESTRUCT_LEVEL;
 
     if(perl_destruct_level < 0) {
 	MP_TRACE_g(fprintf(stderr, 
@@ -269,7 +270,8 @@ void perl_shutdown (server_rec *s, pool *p)
     perl_run_endav("perl_shutdown"); 
 
     MP_TRACE_g(fprintf(stderr, 
-		     "destructing and freeing Perl interpreter..."));
+		     "destructing and freeing Perl interpreter (level=%d)...",
+	       perl_destruct_level));
 
     perl_util_cleanup();
 
@@ -302,11 +304,12 @@ void perl_shutdown (server_rec *s, pool *p)
 
 request_rec *mp_fake_request_rec(server_rec *s, pool *p, char *hook)
 {
-    request_rec *r = (request_rec *)palloc(p, sizeof(request_rec));
+    request_rec *r = (request_rec *)pcalloc(p, sizeof(request_rec));
     r->pool = p; 
     r->server = s;
     r->per_dir_config = NULL;
     r->uri = hook;
+    r->notes = NULL;
     return r;
 }
 
@@ -351,7 +354,7 @@ void perl_restart(server_rec *s, pool *p)
 #endif
 
     /* reload %INC */
-    perl_reload_inc();
+    perl_reload_inc(s, p);
 
     LEAVE;
 
@@ -377,14 +380,14 @@ static void mod_perl_set_cwd(void)
 }
 
 #ifdef PERL_TIE_SCRIPTNAME
-static I32 scriptname_val(IV ix, SV* sv)
+static PERL_MG_UFUNC(scriptname_val, ix, sv)
 { 
     dTHR;
     request_rec *r = perl_request_rec(NULL);
     if(r) 
 	sv_setpv(sv, r->filename);
-    else if(strNE(SvPVX(GvSV(curcop->cop_filegv)), "-e"))
-	sv_setsv(sv, GvSV(curcop->cop_filegv));
+    else if(strNE(SvPVX(GvSV(CopFILEGV(curcop))), "-e"))
+	sv_setsv(sv, GvSV(CopFILEGV(curcop)));
     else {
 	SV *file = perl_eval_pv("(caller())[1]",TRUE);
 	sv_setsv(sv, file);
@@ -411,43 +414,113 @@ static void mod_perl_tie_scriptname(void)
     if(orig_inc) SvREFCNT_dec(orig_inc); \
     orig_inc = av_copy_array(GvAV(incgv))
 
-#if MODULE_MAGIC_NUMBER >= MMN_130
-static void mp_dso_unload(void *data) 
-{ 
-    module *modp;
+#define dl_librefs "DynaLoader::dl_librefs"
+#define dl_modules "DynaLoader::dl_modules"
 
-    if(!PERL_DSO_UNLOAD)
-	return;
+static array_header *xs_dl_librefs(pool *p)
+{
+    I32 i;
+    AV *librefs = perl_get_av(dl_librefs, FALSE);
+    AV *modules = perl_get_av(dl_modules, FALSE);
+    array_header *arr;
 
-    if(strEQ(top_module->name, "mod_perl.c"))
-	return;
+    if (!librefs) {
+	MP_TRACE_g(fprintf(stderr, 
+			   "Could not get @%s for unloading.\n",
+			   dl_librefs));
+	return NULL;
+    }
 
-    for(modp = top_module; modp; modp = modp->next) {
-	if(modp->dynamic_load_handle) {
+    arr = ap_make_array(p, AvFILL(librefs)-1, sizeof(void *));
+
+    for (i=0; i<=AvFILL(librefs); i++) {
+	void *handle;
+	SV *handle_sv = *av_fetch(librefs, i, FALSE);
+	SV *module_sv = *av_fetch(modules, i, FALSE);
+
+	if(!handle_sv) {
 	    MP_TRACE_g(fprintf(stderr, 
-			       "mod_perl: cancel dlclose for %s\n", 
-			       modp->name));
-	    modp->dynamic_load_handle = NULL;
+			       "Could not fetch $%s[%d]!\n",
+			       dl_librefs, (int)i));
+	    continue;
+	}
+	handle = (void *)SvIV(handle_sv);
+
+	MP_TRACE_g(fprintf(stderr, "%s dl handle == 0x%lx\n",
+			   SvPVX(module_sv), (unsigned long)handle));
+	if (handle) {
+	    *(void **)ap_push_array(arr) = handle;
 	}
     }
-} 
-#endif
 
-#define MP_APACHE_VERSION 1.25
+    av_clear(modules);
+    av_clear(librefs);
+
+    return arr;
+}
+
+static void unload_xs_so(array_header *librefs)
+{
+    int i;
+
+    if (!librefs) {
+	return;
+    }
+
+    for (i=0; i < librefs->nelts; i++) {
+	void *handle = ((void **)librefs->elts)[i];
+	MP_TRACE_g(fprintf(stderr, "unload_xs_so: 0x%lx\n",
+			   (unsigned long)handle));
+#ifdef _AIX
+	/* make sure Perl's dlclose is used, instead of Apache's */
+	dlclose(handle);
+#else
+	ap_os_dso_unload(handle);
+#endif
+    }
+}
+
+static void mp_dso_unload(void *data) 
+{ 
+    array_header *librefs;
+
+    librefs = xs_dl_librefs((pool *)data);
+    perl_destruct_level = 2;
+    perl_shutdown(NULL, NULL);
+    unload_xs_so(librefs);
+} 
+
+static void mp_server_notstarting(void *data) 
+{
+    saveINC;
+    require_Apache(NULL); 
+    Apache__ServerStarting(FALSE);
+}
+
+#define Apache__ServerStarting_on() \
+    Apache__ServerStarting(PERL_RUNNING()); \
+    if(!PERL_IS_DSO) \
+        register_cleanup(p, NULL, mp_server_notstarting, mod_perl_noop) 
+
+#define MP_APACHE_VERSION "1.27"
 
 void mp_check_version(void)
 {
     I32 i;
     SV *namesv;
-    SV *version = perl_get_sv("Apache::VERSION", FALSE);
-    if(!version)
+    SV *version;
+    STRLEN n_a;
+
+    require_Apache(NULL);
+
+    if(!(version = perl_get_sv("Apache::VERSION", FALSE)))
 	croak("Apache.pm failed to load!"); /*should never happen*/
-    if(SvNV(version) >= MP_APACHE_VERSION) /*no worries*/
+    if(strEQ(SvPV(version,n_a), MP_APACHE_VERSION)) /*no worries*/
 	return;
 
-    fprintf(stderr, "Apache.pm version %.02f or higher required!\n", 
+    fprintf(stderr, "Apache.pm version %s required!\n", 
 	    MP_APACHE_VERSION);
-    fprintf(stderr, "%s", form("%_ is only version %_\n", 
+    fprintf(stderr, "%s", form("%_ is version %_\n", 
 			       *hv_fetch(GvHV(incgv), "Apache.pm", 9, FALSE),
 			       version));
     fprintf(stderr, 
@@ -469,24 +542,42 @@ void mp_check_version(void)
     exit(1);
 }
 
+#if !HAS_MMN_136
+static void set_sigpipe(void)
+{
+    char *dargs[] = { NULL };
+    perl_require_module("Apache::SIG", NULL);
+    perl_call_argv("Apache::SIG::set", G_DISCARD, dargs);
+}
+#endif
+
+void perl_module_init(server_rec *s, pool *p)
+{
+#if HAS_MMN_130
+    ap_add_version_component(MOD_PERL_STRING_VERSION);
+    if(PERL_RUNNING()) {
+#ifdef PERL_IS_5_6
+	char *version = form("Perl/v%vd", PL_patchlevel);
+#else
+	char *version = form("Perl/%_", perl_get_sv("]", TRUE));
+#endif
+	if(perl_get_sv("Apache::Server::AddPerlVersion", FALSE)) {
+	    ap_add_version_component(version);
+	}
+    }
+#endif
+    perl_startup(s, p);
+}
+
 void perl_startup (server_rec *s, pool *p)
 {
     char *argv[] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
-    char **list, *dstr;
+    char **entries, *dstr;
     int status, i, argc=1;
-    char *dash_e = "BEGIN { $ENV{MOD_PERL} = 1; $ENV{GATEWAY_INTERFACE} = 'CGI-Perl/1.1'; }";
-    char *line_info = "#line 1 mod_perl";
     dPSRV(s);
     SV *pool_rv, *server_rv;
     GV *gv, *shgv;
 
-#if MODULE_MAGIC_NUMBER >= 19980507
-#ifndef MOD_PERL_STRING_VERSION
-#include "mod_perl_version.h"
-#endif
-    ap_add_version_component(MOD_PERL_STRING_VERSION);
-#endif
-
 #ifndef WIN32
     argv[0] = server_argv0;
 #endif
@@ -512,12 +603,21 @@ void perl_startup (server_rec *s, pool *p)
     dstr = NULL;
 #endif
 
+    if(PERL_RUNNING()) {
+	saveINC;
+	mp_check_version();
+#if !HAS_MMN_136
+	set_sigpipe();
+#endif
+    }
+    
     if(perl_is_running == 0) {
 	/* we'll boot Perl below */
     }
     else if(perl_is_running < PERL_DONE_STARTUP) {
 	/* skip the -HUP at server-startup */
 	perl_is_running++;
+	Apache__ServerStarting_on();
 	MP_TRACE_g(fprintf(stderr, "perl_startup: perl aleady running...ok\n"));
 	return;
     }
@@ -543,16 +643,12 @@ void perl_startup (server_rec *s, pool *p)
     if(cls->PerlWarn)
 	argv[argc++] = "-w";
 
-#ifdef PERL_MARK_WHERE
-    argv[argc++] = "-e";
-    argv[argc++] = line_info;
+#ifdef WIN32
+    argv[argc++] = "nul";
 #else
-    line_info = NULL; 
+    argv[argc++] = "/dev/null";
 #endif
 
-    argv[argc++] = "-e";
-    argv[argc++] = dash_e;
-
     MP_TRACE_g(fprintf(stderr, "perl_parse args: "));
     for(i=1; i<argc; i++)
 	MP_TRACE_g(fprintf(stderr, "'%s' ", argv[i]));
@@ -564,7 +660,11 @@ void perl_startup (server_rec *s, pool *p)
 # endif
 #endif
 
+#ifndef perl_init_i18nl10n
     perl_init_i18nl10n(1);
+#else
+    /* 5.6 calls during perl_construct() */
+#endif
 
     MP_TRACE_g(fprintf(stderr, "allocating perl interpreter..."));
     if((perl = perl_alloc()) == NULL) {
@@ -601,17 +701,12 @@ void perl_startup (server_rec *s, pool *p)
     GvSV_setiv(gv, TRUE);
 #endif
 
-    gv = GvSV_init("Apache::__T");
-    if(cls->PerlTaintCheck) 
-	GvSV_setiv(gv, TRUE);
-    SvREADONLY_on(GvSV(gv));
-
+    perl_tainting_set(s, cls->PerlTaintCheck);
     (void)GvSV_init("Apache::__SendHeader");
     (void)GvSV_init("Apache::__CurrentCallback");
-    (void)GvHV_init("mod_perl::UNIMPORT");
 
     Apache__ServerReStarting(FALSE); /* just for -w */
-    Apache__ServerStarting(PERL_RUNNING());
+    Apache__ServerStarting_on();
 
 #ifdef PERL_STACKED_HANDLERS
     if(!stacked_handlers) {
@@ -631,13 +726,32 @@ void perl_startup (server_rec *s, pool *p)
     }
     MP_TRACE_g(fprintf(stderr, "ok\n"));
 
+    /* Force the environment to be copied out of its original location
+       above argv[].  This fixes a crash caused when a module called putenv()
+       before any Perl modified the environment - environ would change to a
+       new value, and the check in my_setenv() to duplicate the environment
+       would fail, and then setting some environment value which had a previous
+       value would cause perl to try to free() something from the original env.
+       This crashed free(). */
+    my_setenv("MODPERL_ENV_FIXUP", "0");
+    my_setenv("MODPERL_ENV_FIXUP", NULL);
+
     {
 	dTHR;
 	TAINT_NOT; /* At this time all is safe */
     }
 
+#ifdef MOD_PERL_PREFIX
+	av_unshift(GvAV(incgv),1);
+	av_store(GvAV(incgv), 0, newSVpv(MOD_PERL_PREFIX,0));
+#endif
+	
+#ifdef APACHE_PERL5LIB
+    perl_incpush(APACHE_PERL5LIB);
+#else
     av_push(GvAV(incgv), newSVpv(server_root_relative(p,""),0));
     av_push(GvAV(incgv), newSVpv(server_root_relative(p,"lib/perl"),0));
+#endif
 
     /* *CORE::GLOBAL::exit = \&Apache::exit */
     if(gv_stashpv("CORE::GLOBAL", FALSE)) {
@@ -646,36 +760,27 @@ void perl_startup (server_rec *s, pool *p)
 	GvIMPORTED_CV_on(exitgp);
     }
 
-    if(PERL_STARTUP_DONE_CHECK && !getenv("PERL_STARTUP_DONE")) {
-	MP_TRACE_g(fprintf(stderr, 
-			   "mod_perl: PerlModule,PerlRequire postponed\n"));
-	my_setenv("PERL_STARTUP_DONE", "1");
-	saveINC;
-	Apache__ServerStarting(FALSE);
-	return;
-    }
-
     ENTER_SAFE(s,p);
     MP_TRACE_g(mod_perl_dump_opmask());
 
-    list = (char **)cls->PerlRequire->elts;
+    entries = (char **)cls->PerlRequire->elts;
     for(i = 0; i < cls->PerlRequire->nelts; i++) {
-	if(perl_load_startup_script(s, p, list[i], TRUE) != OK) {
+	if(perl_load_startup_script(s, p, entries[i], TRUE) != OK) {
 	    fprintf(stderr, "Require of Perl file `%s' failed, exiting...\n", 
-		    list[i]);
+		    entries[i]);
 	    exit(1);
 	}
     }
 
-    list = (char **)cls->PerlModule->elts;
+    entries = (char **)cls->PerlModule->elts;
     for(i = 0; i < cls->PerlModule->nelts; i++) {
-	if(perl_require_module(list[i], s) != OK) {
+	if(perl_require_module(entries[i], s) != OK) {
 	    fprintf(stderr, "Can't load Perl module `%s', exiting...\n", 
-		    list[i]);
+		    entries[i]);
 	    exit(1);
 	}
     }
-    mp_check_version();
+
     LEAVE_SAFE;
 
     MP_TRACE_g(fprintf(stderr, 
@@ -687,17 +792,19 @@ void perl_startup (server_rec *s, pool *p)
 #endif
 
     saveINC;
-    Apache__ServerStarting(FALSE);
-#if MODULE_MAGIC_NUMBER >= MMN_130
-    if(perl_module.dynamic_load_handle) 
-	register_cleanup(p, NULL, mp_dso_unload, NULL); 
-#endif
+
+    if (PERL_IS_DSO) {
+	register_cleanup(p, p, mp_dso_unload, null_cleanup);
+    }
 }
 
 int mod_perl_sent_header(request_rec *r, int val)
 {
     dPPDIR;
 
+    if (val == DONE) {
+        val = r->assbackwards = 1; /* so apache does not send another header */
+    }
     if(val) MP_SENTHDR_on(cld);
     val = MP_SENTHDR(cld) ? 1 : 0;
     return MP_SENDHDR(cld) ? val : 1;
@@ -711,11 +818,23 @@ int perl_handler(request_rec *r)
 {
     dSTATUS;
     dPPDIR;
+    dPPREQ;
     dTHR;
-    SV *nwvh = Nullsv;
+    GV *gv;
+
+#ifdef USE_ITHREADS
+    dTHX;
+
+    if (!aTHX) {
+        PERL_SET_CONTEXT(perl);
+    }
+#endif
 
     (void)acquire_mutex(mod_perl_mutex);
-    
+
+    gv = gv_fetchpv("SIG", TRUE, SVt_PVHV);
+
+   
 #if 0
     /* force 'PerlSendHeader On' for sub-requests
      * e.g. Apache::Sandwich 
@@ -727,8 +846,6 @@ int perl_handler(request_rec *r)
     if(MP_SENDHDR(cld)) 
 	MP_SENTHDR_off(cld);
 
-    table_set(r->subprocess_env, "MOD_PERL", MOD_PERL_VERSION);
-
     (void)perl_request_rec(r); 
 
     MP_TRACE_g(fprintf(stderr, "perl_handler ENTER: SVs = %5d, OBJs = %5d\n",
@@ -736,32 +853,40 @@ int perl_handler(request_rec *r)
     ENTER;
     SAVETMPS;
 
-    if((nwvh = ApachePerlRun_name_with_virtualhost())) {
-	if(!r->server->is_virtual) {
-	    SAVESPTR(nwvh);
-	    sv_setiv(nwvh, 0);
-	}
+    if (gv) {
+	save_hptr(&GvHV(gv)); 
     }
 
-    save_hptr(&GvHV(siggv)); 
-
-    save_aptr(&endav); 
-    endav = Nullav;
+    if (endav) {
+	save_aptr(&endav); 
+	endav = Nullav;
+    }
 
     /* hookup STDIN & STDOUT to the client */
     perl_stdout2client(r);
     perl_stdin2client(r);
 
-    if(MP_ENV(cld)) 
-	perl_setup_env(r);
+    if(!cfg) {
+        cfg = perl_create_request_config(r->pool, r->server);
+        set_module_config(r->request_config, &perl_module, cfg);
+    }
 
+    cfg->setup_env = 1;
     PERL_CALLBACK("PerlHandler", cld->PerlHandler);
+    cfg->setup_env = 0;
 
     FREETMPS;
     LEAVE;
     MP_TRACE_g(fprintf(stderr, "perl_handler LEAVE: SVs = %5d, OBJs = %5d\n", 
 		     (int)sv_count, (int)sv_objcount));
 
+    if (r->prev && (r->prev->status != HTTP_OK) &&
+        mod_perl_sent_header(r, 0))
+    {
+        /* avoid recursive error for ErrorDocuments */
+        status = OK;
+    }
+
     (void)release_mutex(mod_perl_mutex);
     return status;
 }
@@ -793,6 +918,7 @@ void PERL_CHILD_INIT_HOOK(server_rec *s, pool *p)
     register_cleanup(p, args, perl_child_exit_cleanup, null_cleanup);
 
     mod_perl_init_ids();
+    Apache__ServerStarting(FALSE);
     PERL_CALLBACK(hook, cls->PerlChildInitHandler);
 }
 #endif
@@ -811,17 +937,30 @@ void PERL_CHILD_EXIT_HOOK(server_rec *s, pool *p)
 }
 #endif
 
+static int do_proxy (request_rec *r)
+{
+    return r->parsed_uri.scheme &&
+	!(r->parsed_uri.hostname
+	  && strEQ(r->parsed_uri.scheme, ap_http_method(r))
+	  && ap_matches_request_vhost(r, r->parsed_uri.hostname,
+				      r->parsed_uri.port_str ? 
+				      r->parsed_uri.port : 
+				      ap_default_port(r)));
+}
+
 #ifdef PERL_POST_READ_REQUEST
 int PERL_POST_READ_REQUEST_HOOK(request_rec *r)
 {
     dSTATUS;
     dPSRV(r->server);
+#ifdef PERL_TRANS
 #if MODULE_MAGIC_NUMBER > 19980270
-    if(r->parsed_uri.scheme && r->parsed_uri.hostname) {
+    if (cls->PerlTransHandler && do_proxy(r)) {
 	r->proxyreq = 1;
 	r->uri = r->unparsed_uri;
     }
 #endif
+#endif
 #ifdef PERL_INIT
     PERL_CALLBACK("PerlInitHandler", cls->PerlInitHandler);
 #endif
@@ -915,14 +1054,54 @@ int PERL_LOG_HOOK(request_rec *r)
 }
 #endif
 
-#define CleanupHandler cld->PerlCleanupHandler
-
 #ifdef PERL_STACKED_HANDLERS
-#define has_CleanupHandler (CleanupHandler && SvREFCNT(CleanupHandler))
+#define CleanupHandler \
+((cld->PerlCleanupHandler && SvREFCNT(cld->PerlCleanupHandler)) ? cld->PerlCleanupHandler : Nullav)
 #else
-#define has_CleanupHandler CleanupHandler
+#define CleanupHandler cld->PerlCleanupHandler
 #endif
 
+#ifdef PERL_TRACE
+static char *my_signame(I32 num)
+{
+#ifdef psig_name
+    return Perl_psig_name[num] ?
+	SvPV(Perl_psig_name[num],na) : "?";
+#else
+    return PL_sig_name[num];
+#endif
+}
+
+#endif
+
+static void per_request_cleanup(request_rec *r)
+{
+    dPPREQ;
+    perl_request_sigsave **sigs;
+    int i;
+
+    if(!cfg) {
+	return;
+    }
+    if(cfg->pnotes) {
+	hv_clear(cfg->pnotes);
+	SvREFCNT_dec(cfg->pnotes);
+	cfg->pnotes = Nullhv;
+    }
+
+#ifndef WIN32
+    sigs = (perl_request_sigsave **)cfg->sigsave->elts;
+    for (i=0; i < cfg->sigsave->nelts; i++) {
+	MP_TRACE_g(fprintf(stderr, 
+			   "mod_perl: restoring SIG%s (%d) handler from: 0x%lx to: 0x%lx\n",
+			   my_signame(sigs[i]->signo), (int)sigs[i]->signo,
+			   (unsigned long)rsignal_state(sigs[i]->signo),
+			   (unsigned long)sigs[i]->h));
+	rsignal(sigs[i]->signo, sigs[i]->h);
+    }
+#endif
+}
+
 void mod_perl_end_cleanup(void *data)
 {
     request_rec *r = (request_rec *)data;
@@ -930,13 +1109,12 @@ void mod_perl_end_cleanup(void *data)
     dPPDIR;
 
 #ifdef PERL_CLEANUP
-    if(has_CleanupHandler) {
-	PERL_CALLBACK("PerlCleanupHandler", cld->PerlCleanupHandler);
-    }
+    PERL_CALLBACK("PerlCleanupHandler", CleanupHandler);
 #endif
 
     MP_TRACE_g(fprintf(stderr, "perl_end_cleanup..."));
     perl_run_rgy_endav(r->uri);
+    per_request_cleanup(r);
 
     /* clear %ENV */
     perl_clear_env();
@@ -948,7 +1126,7 @@ void mod_perl_end_cleanup(void *data)
     GvAV(incgv) = av_copy_array(orig_inc);
 
     /* reset $/ */
-    sv_setpvn(GvSV(gv_fetchpv("/", FALSE, SVt_PV)), "\n", 1);
+    sv_setpvn(GvSV(gv_fetchpv("/", TRUE, SVt_PV)), "\n", 1);
 
     {
 	dTHR;
@@ -960,17 +1138,17 @@ void mod_perl_end_cleanup(void *data)
 
 #ifdef PERL_STACKED_HANDLERS
     /* reset Apache->push_handlers, but don't clear ExitHandler */
-#define CH_EXIT_KEY "PerlChildExitHandler", 20
+#define CH_EXIT_KEY "PerlChildExitHandler"
     {
 	SV *exith = Nullsv;
-	if(hv_exists(stacked_handlers, CH_EXIT_KEY)) {
-	    exith = *hv_fetch(stacked_handlers, CH_EXIT_KEY, FALSE);
+	if(hv_exists(stacked_handlers, CH_EXIT_KEY, 20)) {
+	    exith = *hv_fetch(stacked_handlers, CH_EXIT_KEY, 20, FALSE);
             /* inc the refcnt since hv_clear will dec it */
 	    ++SvREFCNT(exith);
 	}
 	hv_clear(stacked_handlers);
 	if(exith) 
-	    hv_store(stacked_handlers, CH_EXIT_KEY, exith, FALSE);
+	    hv_store(stacked_handlers, CH_EXIT_KEY, 20, exith, FALSE);
     }
 
 #endif
@@ -985,7 +1163,7 @@ void mod_perl_end_cleanup(void *data)
 
 void mod_perl_cleanup_handler(void *data)
 {
-    request_rec *r = perl_request_rec(NULL);
+    request_rec *r = (request_rec *)data;
     SV *cv;
     I32 i;
     dPPDIR;
@@ -1006,7 +1184,7 @@ void mod_perl_cleanup_handler(void *data)
 }
 
 #ifdef PERL_METHOD_HANDLERS
-int perl_handler_ismethod(HV *class, char *sub)
+int perl_handler_ismethod(HV *pclass, char *sub)
 {
     CV *cv;
     HV *stash;
@@ -1017,12 +1195,19 @@ int perl_handler_ismethod(HV *class, char *sub)
     if(!sub) return 0;
     sv = newSVpv(sub,0);
     if(!(cv = sv_2cv(sv, &stash, &gv, FALSE))) {
-	GV *gvp = gv_fetchmethod(class, sub);
+	GV *gvp = gv_fetchmethod(pclass, sub);
 	if (gvp) cv = GvCV(gvp);
     }
 
-    if (cv && SvPOK(cv)) 
+#ifdef CVf_METHOD
+    if (cv && (CvFLAGS(cv) & CVf_METHOD)) {
+        is_method = 1;
+    }
+#endif
+    if (!is_method && (cv && SvPOK(cv))) {
 	is_method = strnEQ(SvPVX(cv), "$$", 2);
+    }
+
     MP_TRACE_h(fprintf(stderr, "checking if `%s' is a method...%s\n", 
 	   sub, (is_method ? "yes" : "no")));
     SvREFCNT_dec(sv);
@@ -1105,6 +1290,14 @@ int perl_run_stacked_handlers(char *hook, request_rec *r, AV *handlers)
     SV *sub, **svp; 
     int hook_len = strlen(hook);
 
+#ifdef USE_ITHREADS
+    dTHX;
+
+    if (!aTHX) {
+        PERL_SET_CONTEXT(perl);
+    }
+#endif
+
     if(handlers == Nullav) {
 	if(hv_exists(stacked_handlers, hook, hook_len)) {
 	   svp = hv_fetch(stacked_handlers, hook, hook_len, 0);
@@ -1125,7 +1318,12 @@ int perl_run_stacked_handlers(char *hook, request_rec *r, AV *handlers)
       /* XXX: bizarre, 
 	 I only see this with httpd.conf.pl and PerlAccessHandler */
 	if(SvTYPE((SV*)handlers) != SVt_PVAV) {
-	    fprintf(stderr, "[warning] %s stack is not an ARRAY!\n", hook);
+#if MODULE_MAGIC_NUMBER > 19970909 
+	    aplog_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r->server,
+#else
+	    fprintf(stderr, 
+#endif
+		    "[warning] %s stack is not an ARRAY!\n", hook);
 	    sv_dump((SV*)handlers);
 	    return DECLINED;
 	}
@@ -1170,10 +1368,8 @@ int perl_run_stacked_handlers(char *hook, request_rec *r, AV *handlers)
 void perl_per_request_init(request_rec *r)
 {
     dPPDIR;
-
-    /* PerlSetEnv */
-    mod_perl_dir_env(cld);
-
+    dPPREQ;
+    
     /* PerlSendHeader */
     if(MP_SENDHDR(cld)) {
 	MP_SENTHDR_off(cld);
@@ -1183,17 +1379,43 @@ void perl_per_request_init(request_rec *r)
     else
 	MP_SENTHDR_on(cld);
 
+    if(!cfg) {
+	cfg = perl_create_request_config(r->pool, r->server);
+	set_module_config(r->request_config, &perl_module, cfg);
+    }
+    else if (cfg->setup_env && MP_ENV(cld)) { 
+	perl_setup_env(r);
+	cfg->setup_env = 0; /* just once per-request */
+    }
+
+    if (cfg->dir_env != cld->env) {
+        /* PerlSetEnv
+         * update only if the table changes across a request
+         */
+        mod_perl_dir_env(r, cld);
+        cfg->dir_env = cld->env;
+    }
+
+    if(callbacks_this_request++ > 0) return;
+
+    if (!r->main) {
+	/* so Apache->request will work before PerlHandler with CGI.pm
+	 * XXX: triggers core dump in subrequests, 
+	 * so just do in the main request for now
+	 */
+	(void)perl_request_rec(r);
+    }
+
     /* SetEnv PERL5LIB */
-    if(!MP_INCPUSH(cld)) {
+    if (!MP_INCPUSH(cld)) {
 	char *path = (char *)table_get(r->subprocess_env, "PERL5LIB");
-	if(path) {
+
+	if (path) {
 	    perl_incpush(path);
 	    MP_INCPUSH_on(cld);
 	}
     }
 
-    if(callbacks_this_request++ > 0) return;
-
     {
 	dPSRV(r->server);
 	mod_perl_pass_env(r->pool, cls);
@@ -1229,14 +1451,14 @@ int perl_call_handler(SV *sv, request_rec *r, AV *args)
     dSP;
     perl_dir_config *cld = NULL;
     HV *stash = Nullhv;
-    SV *class = newSVsv(sv), *dispsv = Nullsv;
+    SV *pclass = newSVsv(sv), *dispsv = Nullsv;
     CV *cv = Nullcv;
     char *method = "handler";
     int defined_sub = 0, anon = 0;
     char *dispatcher = NULL;
 
     if(r->per_dir_config)
-	cld = get_module_config(r->per_dir_config, &perl_module);
+	cld = (perl_dir_config *) get_module_config(r->per_dir_config, &perl_module);
 
 #ifdef PERL_DISPATCH
     if(cld && (dispatcher = cld->PerlDispatchHandler)) {
@@ -1255,7 +1477,7 @@ int perl_call_handler(SV *sv, request_rec *r, AV *args)
 	perl_per_request_init(r);
 
     if(!dispatcher && (SvTYPE(sv) == SVt_PV)) {
-	char *imp = pstrdup(r->pool, (char *)SvPV(class,na));
+	char *imp = pstrdup(r->pool, (char *)SvPV(pclass,na));
 
 	if((anon = strnEQ(imp,"sub ",4))) {
 	    sv = perl_eval_pv(imp, FALSE);
@@ -1267,37 +1489,37 @@ int perl_call_handler(SV *sv, request_rec *r, AV *args)
 
 #ifdef PERL_METHOD_HANDLERS
 	{
-	    char *end_class = NULL;
+	    char *end_pclass = NULL;
 
-	    if ((end_class = strstr(imp, "->"))) {
-		end_class[0] = '\0';
-		if(class)
-		    SvREFCNT_dec(class);
-		class = newSVpv(imp, 0);
-		end_class[0] = ':';
-		end_class[1] = ':';
-		method = &end_class[2];
+	    if ((end_pclass = strstr(imp, "->"))) {
+		end_pclass[0] = '\0';
+		if(pclass)
+		    SvREFCNT_dec(pclass);
+		pclass = newSVpv(imp, 0);
+		end_pclass[0] = ':';
+		end_pclass[1] = ':';
+		method = &end_pclass[2];
 		imp = method;
 		++is_method;
 	    }
 	}
 
-	if(*SvPVX(class) == '$') {
-	    SV *obj = perl_eval_pv(SvPVX(class), TRUE);
+	if(*SvPVX(pclass) == '$') {
+	    SV *obj = perl_eval_pv(SvPVX(pclass), TRUE);
 	    if(SvROK(obj) && sv_isobject(obj)) {
 		MP_TRACE_h(fprintf(stderr, "handler object %s isa %s\n",
-				   SvPVX(class),  HvNAME(SvSTASH((SV*)SvRV(obj)))));
-		SvREFCNT_dec(class);
-		class = obj;
-		++SvREFCNT(class); /* this will _dec later */
-		stash = SvSTASH((SV*)SvRV(class));
+				   SvPVX(pclass),  HvNAME(SvSTASH((SV*)SvRV(obj)))));
+		SvREFCNT_dec(pclass);
+		pclass = obj;
+		++SvREFCNT(pclass); /* this will _dec later */
+		stash = SvSTASH((SV*)SvRV(pclass));
 	    }
 	}
 
-	if(class && !stash) stash = gv_stashpv(SvPV(class,na),FALSE);
+	if(pclass && !stash) stash = gv_stashpv(SvPV(pclass,na),FALSE);
 	   
 #if 0
-	MP_TRACE_h(fprintf(stderr, "perl_call: class=`%s'\n", SvPV(class,na)));
+	MP_TRACE_h(fprintf(stderr, "perl_call: pclass=`%s'\n", SvPV(pclass,na)));
 	MP_TRACE_h(fprintf(stderr, "perl_call: imp=`%s'\n", imp));
 	MP_TRACE_h(fprintf(stderr, "perl_call: method=`%s'\n", method));
 	MP_TRACE_h(fprintf(stderr, "perl_call: stash=`%s'\n", 
@@ -1345,6 +1567,7 @@ int perl_call_handler(SV *sv, request_rec *r, AV *args)
 	    if(stash) /* check again */
 		is_method = perl_handler_ismethod(stash, method);
 #endif
+	    SPAGAIN; /* reset stack pointer after require() */
 	}
 	
 	if(!is_method && !defined_sub) {
@@ -1374,11 +1597,11 @@ callback:
     PUSHMARK(sp);
 #ifdef PERL_METHOD_HANDLERS
     if(is_method)
-	XPUSHs(sv_2mortal(class));
+	XPUSHs(sv_2mortal(pclass));
     else
-	SvREFCNT_dec(class);
+	SvREFCNT_dec(pclass);
 #else
-    SvREFCNT_dec(class);
+    SvREFCNT_dec(pclass);
 #endif
 
     XPUSHs((SV*)perl_bless_request_rec(r)); 
@@ -1413,10 +1636,17 @@ callback:
     
     SPAGAIN;
 
-    if(perl_eval_ok(r->server) != OK) {
-	MP_STORE_ERROR(r->uri, ERRSV);
-	if(!perl_sv_is_http_code(ERRSV, &status))
-	    status = SERVER_ERROR;
+    if ((status = perl_eval_ok(r->server)) != OK) {
+        dTHRCTX;
+        if (status == SERVER_ERROR) {
+            MP_STORE_ERROR(r->uri, ERRSV);
+            if (r->notes) {
+                ap_table_set(r->notes, "error-notes", SvPVX(ERRSV));
+            }
+        }
+        else if (status == DECLINED) {
+            status = r->status == 200 ? OK : r->status;
+        }
     }
     else if(count != 1) {
 	mod_perl_error(r->server,
@@ -1443,8 +1673,11 @@ callback:
     MP_TRACE_g(fprintf(stderr, "perl_call_handler: SVs = %5d, OBJs = %5d\n", 
 	    (int)sv_count, (int)sv_objcount));
 
-    if(SvMAGICAL(ERRSV))
-       sv_unmagic(ERRSV, 'U'); /* Apache::exit was called */
+    {
+	dTHRCTX;
+	if(SvMAGICAL(ERRSV))
+	    sv_unmagic(ERRSV, 'U'); /* Apache::exit was called */
+    }
 
     return status;
 }
